import cosmotool as ct
import numpy as np
import healpy as hp

d = np.zeros((64,64,64), ct.DTYPE)

d[32,32,32] = 1

ii=np.arange(256)*64/256.-32
xx = ii[:,None,None].repeat(256,axis=1).repeat(256,axis=2).reshape(256**3)
yy = ii[None,:,None].repeat(256,axis=0).repeat(256,axis=2).reshape(256**3)
zz = ii[None,None,:].repeat(256,axis=0).repeat(256,axis=1).reshape(256**3)

d_high = ct.interp3d(xx, yy, zz, d, 64, periodic=True)
d_high = d_high.reshape((256,256,256))

proj0 = ct.spherical_projection(64, d, 0, 20, integrator_id=0, shifter=np.array([0.5,0.5,0.5]))
proj1 = ct.spherical_projection(64, d, 0, 20, integrator_id=1)

proj0_high = ct.spherical_projection(256, d_high, 0, 30, integrator_id=0, shifter=np.array([0.5,0.5,0.5]))



import numpy as np
import numexpr as ne
from .constants import *

# -----------------------------------------------------------------------------
# Generic profile generator
# -----------------------------------------------------------------------------

class KSZ_Profile(object):
  R_star= 0.0 # 15 kpc/h 
  L_gal0 = 10**(0.4*(tmpp_cat['Msun']-tmpp_cat['Mstar']))

  def __init__(self,sculpt):
    """Base class for KSZ profiles

    Arguments:
      sculpt (float): If negative, do not sculpt. If positive, there will be a 2d 
                      suppression of the profile with a radius given by sculpt (in arcmins).
    """
    self.sculpt = sculpt * np.pi/180/60.
    self.rGalaxy = 1.0

  def evaluate_profile(self, r):
    raise NotImplementedError("Abstract function")

  def projected_profile(self, cos_theta,angularDistance):

    idx_base = idx = np.where(cos_theta > 0)[0]
    tan_theta_2 = 1/(cos_theta[idx]**2) - 1
    tan_theta_2_max = (self.rGalaxy/angularDistance)**2
    tan_theta_2_min = (self.R_star/angularDistance)**2

    idx0 = np.where((tan_theta_2 < tan_theta_2_max))
    idx = idx_base[idx0]
    tan_theta_2 = tan_theta_2[idx0]
    tan_theta = np.sqrt(tan_theta_2)

    r = (tan_theta*angularDistance)

    m,idx_mask = self.evaluate_profile(r)
    idx_mask = idx[idx_mask]

    idx_mask = np.append(idx_mask,idx[np.where(tan_theta_2<tan_theta_2_min)[0]])
    if tan_theta_2.size > 0:
      idx_mask = np.append(idx_mask,idx[tan_theta_2.argmin()])

    if self.sculpt > 0:
      theta = np.arctan(tan_theta)
      cond = theta < self.sculpt
      m[cond] *= (theta[cond]/self.sculpt)**2

    return idx,idx_mask,m
     

# -----------------------------------------------------------------------------
# Isothermal profile generator
# -----------------------------------------------------------------------------

class KSZ_Isothermal(KSZ_Profile):
  sigma_FP=160e3  #m/s
  R_innergal = 0.030

  def __init__(self, Lgal, x, y=0.0, sculpt=-1):
    """Support for Isothermal profile

    Arguments:
      Lgal (float): Galaxy luminosity in solar units
      x (float): extent of halo in virial radius units

    Keyword arguments:
      y (float): Inner part where there is no halo
      sculpt (float): If negative, do not sculpt. If positive, there will be a 2d 
                      suppression of the profile with a radius given by sculpt (in arcmins).
    """
    
    super(KSZ_Isothermal,self).__init__(sculpt)
    
    self.R_gal = 0.226 * x
    self.R_innergal *= y

    self.rho0 = self.sigma_FP**2/(2*np.pi*G) # * (Lgal/L_gal0)**(2./3)
    self.rGalaxy = self.R_gal*(Lgal/self.L_gal0)**(1./3)
    self.rInnerGalaxy = self.R_innergal*(Lgal/self.L_gal0)**(1./3)
    self._prepare()
    
  def _prepare(self):
    pass

  def evaluate_profile(self,r):
    rho0, rGalaxy, rInner = self.rho0, self.rGalaxy, self.rInnerGalaxy
    
    D = {'rho0':rho0, 'rGalaxy':rGalaxy, 'rInner': rInner, 'Mpc':Mpc }
    
    Q = np.zeros(r.size)

    cond = (r<=1e-10)
    Q[cond] = rho0*2/Mpc * (rGalaxy-rInner)/(rGalaxy*rInner)

    cond = (r>0)*(r <= rInner)
    D['r'] = r[cond]
    Q[cond] = ne.evaluate('rho0*2/(Mpc*r) * (arctan(sqrt( (rGalaxy/r)**2 -1 ))  - arctan(sqrt( (rInner/r)**2 - 1 )))',
                local_dict=D)
 
    cond = (r > rInner)*(r <= rGalaxy)
    D['r'] = r[cond]
    Q[cond] = ne.evaluate('rho0*2/(Mpc*r) * arctan(sqrt( (rGalaxy/r)**2 -1 ))',
                local_dict=D)

    return Q,[] #np.where(r<rInner)[0]


# -----------------------------------------------------------------------------
# NFW profile generator
# -----------------------------------------------------------------------------

class KSZ_NFW(KSZ_Profile):
  """ Support for NFW profile
"""

  def __init__(self,x,y=0.0):
    from numpy import log, pi

    if 'pre_nfw' not in self:
      self._prepare()

    kiso = KSZ_Isothermal(x,y)
    r_is = kiso.rGalaxy
    rho_is = kiso.rho0
    r_inner = kiso.rInnerGalaxy

    self.Mgal = rho_is*4*pi*(r_is/args.x)*Mpc #Lgal*M_over_L_galaxy
    self.Rvir = r_is/x

    cs = self._get_concentration(Mgal)

    self.rs = Rvir/cs
    b = (log(1.+cs)-cs/(1.+cs))
    
    self.rho_s = Mgal/(4*pi*b*(rs*Mpc)**3)


  def _prepare(self, _x_min=1e-4, _x_max=1e4):
    from scipy.integrate import quad
    from numpy import sqrt, log10
    from scipy.interpolate import interp1d

    lmin = log10(x_min)
    lmax = log10(x_max)

    x = 10**(np.arange(100)*(lmax-lmin)/100.+lmin)
    profile = np.empty(x.size)
    
    nu_tilde = lambda u: (1/(u*(1+u)**2))

    for i in range(x.size):
        if x[i] < args.x:
            profile[i] = 2*quad(lambda y: (nu_tilde(sqrt(x[i]**2+y**2))), 0, np.sqrt((args.x)**2-x[i]**2))[0]
        else:
            profile[i] = 0

    # Insert the interpolator into the class definition
    KSZ_NFW.pre_nfw = self.pre_nfw = interp1d(x,prof)

  def _get_concentration(self, Mvir):
  	from numpy import exp, log

  	return exp(0.971 - 0.094*log(Mvir/(1e12*MassSun)))

  def evaluate_profile(self,r):
    cs = self._get_concentration(self.Mvir)
    rs = self.Rvir/cs
    
    return self.rho_s*rs*Mpc*self.pre_nfw(r/rs),np.array([],dtype=int)





import numpy as np

Mpc=3.08e22
rhoc = 1.8864883524081933e-26  # m^(-3)
sigmaT = 6.6524e-29
mp = 1.6726e-27
lightspeed = 299792458.
v_unit = 1e3  # Unit of 1 km/s
T_cmb=2.725
h = 0.71
Y = 0.245  #The Helium abundance
Omega_matter = 0.26
Omega_baryon=0.0445

G=6.67e-11
MassSun=2e30
frac_electron = 1.0 # Hmmmm
frac_gas_galaxy = 0.14
mu = 1/(1-0.5*Y)

tmpp_cat={'Msun':3.29,
    'alpha':-0.7286211634758224,
    'Mstar':-23.172904033796893,
    'PhiStar':0.0113246633636846,
    'lbar':393109973.22508669}

baryon_fraction = Omega_baryon / Omega_matter

ksz_normalization = -T_cmb*sigmaT*v_unit/(lightspeed*mu*mp) * baryon_fraction
assert ksz_normalization < 0
rho_mean_matter = Omega_matter * (3*(100e3/Mpc)**2/(8*np.pi*G)) 
Lbar = tmpp_cat['lbar'] / Mpc**3
M_over_L_galaxy = rho_mean_matter / Lbar


del np



from libcpp cimport bool
from libcpp cimport string as cppstring
from libcpp.vector cimport vector as cppvector
from cython.parallel cimport prange
import numpy as np
cimport numpy as np
from cpython cimport PyObject, Py_INCREF
cimport cython

np.import_array()

cdef extern from "sys/types.h":
   ctypedef np.int64_t int64_t

cdef extern from "loadSimu.hpp" namespace "CosmoTool":

  cdef cppclass SimuData:
    np.float_t BoxSize
    np.float_t time
    np.float_t Hubble

    np.float_t Omega_M
    np.float_t Omega_Lambda
    np.int64_t TotalNumPart
    np.int64_t NumPart
    int64_t *Id
    float *Pos[3]
    float *Vel[3]
    int *type
    float *Mass

    bool noAuto

  cdef const int NEED_GADGET_ID 
  cdef const int NEED_POSITION
  cdef const int NEED_VELOCITY
  cdef const int NEED_TYPE
  cdef const int NEED_MASS

cdef extern from "loadGadget.hpp" namespace "CosmoTool":

  SimuData *loadGadgetMulti(const char *fname, int id, int flags, int gformat) nogil except +
  void cxx_writeGadget "CosmoTool::writeGadget" (const char * s, SimuData *data) except +

cdef extern from "safe_gadget.hpp":
  SimuData *loadGadgetMulti_safe(cppstring.string s, int flags, int gformat) nogil
  SimuData **alloc_simudata(int num) nogil
  void del_simudata(SimuData **d) nogil

cdef extern from "cppHelper.hpp":
  int customCosmotoolHandler() nogil

cdef extern from "loadRamses.hpp" namespace "CosmoTool":
  SimuData *loadRamsesSimu(const char *basename, int id, int cpuid, bool dp, int flags) except +customCosmotoolHandler

class PySimulationBase(object):
  """
  This is the base class to representation Simulation in CosmoTool/python.
  """
  
  def getPositions(self):
    """
    getPositions(self)
    
    Returns:
        A list of three arrays holding the positions of the particles. 
        The i-th element is the i-th coordinate of each particle.
        It may be None if the positions were not requested.
    """
    raise NotImplementedError("getPositions is not implemented")
    
  def getVelocities(self):
    """
    getVelocities(self)
    
    Returns:
        A list of three arrays holding the velocities of the particles. 
        The i-th element is the i-th coordinate of each particle.
        It may be None if the velocities were not requested.
    """
    raise NotImplementedError("getVelocities is not implemented")
    
  def getIdentifiers(self):
    """
    getIdentifiers(self)
    
    Returns:
        Returns an integer array that hold the unique identifiers of
        each particle. 
        It may be None if the identifiers were not requested.
    """
    raise NotImplementedError("getIdentifiers is not implemented")

  def getTypes(self):
    """
    getTypes(self)
    
    Returns:
        Returns an integer array that hold the type of
        each particle. 
        It may be None if the types were not requested.
    """
    raise NotImplementedError("getTypes is not implemented")

  def getOmega_M(self):
    """
    getOmega_M(self)
    
    Returns:
        the mean matter density in the simulation, with respect
        to the critical density.
    """ 
    raise NotImplementedError("getOmega_M is not implemented")
  
  def getOmega_Lambda(self):
    """
    getOmega_Lambda(self)
    
    Returns:
        the mean dark energy density in the simulation, with respect
        to the critical density.
    """ 
    raise NotImplementedError("getOmega_Lambda is not implemented")

  def getTime(self):
    """
    getTime(self)
    
    Returns:
        the time the snapshot was taken in the simulation. It can
        have various units depending on the file format.
    """
    raise NotImplementedError("getTime is not implemented")

  def getHubble(self):
    """
    getHubble(self)
    
    Returns:
        the hubble constant in unit of 100 km/s/Mpc
    """
    raise NotImplementedError("getHubble is not implemented")

  def getBoxsize(self):
    """
    getBoxsize(self)
    
    Returns:
        the size of the simulation box. The length unit is not fixed,
        though it is customary to have it in Mpc/h if the loader has
        access to the unit normalization.
    """
    raise NotImplementedError("getBoxsize is not implemented")

  def getMasses(self):
    """
    getMasses(self)
    
    Returns:
        an array with the masses of each particles, in unspecified unit that
        depend on the loader.
    """
    raise NotImplementedError("getMasses is not implemented")

cdef class Simulation:
    """
    Simulation()
    
    Class that directly manages internal loaded data obtained from a loader
    """

    cdef list positions
    cdef list velocities
    cdef object identifiers
    cdef object types
    cdef object masses

    cdef SimuData *data

    property BoxSize:
        def __get__(Simulation self):
            return self.data.BoxSize
        
    property time:
        def __get__(Simulation self):
            return self.data.time

    property Hubble:
        def __get__(Simulation self):
            return self.data.Hubble

    property Omega_M:
        def __get__(Simulation self):
            return self.data.Omega_M
        
    property Omega_Lambda:
        def __get__(Simulation self):
            return self.data.Omega_Lambda
        
    property positions:
        def __get__(Simulation self):
            return self.positions
        
    property velocities:
        def __get__(Simulation self):
            return self.velocities

    property identifiers:
        def __get__(Simulation self):
            return self.identifiers

    property types:
        def __get__(Simulation self):
            return self.types

    property masses:
        def __get__(Simulation self):
            return self.masses
        
    property numParticles:
        def __get__(Simulation self):
            return self.data.NumPart


    property totalNumParticles:
        def __get__(Simulation self):
            return self.data.TotalNumPart

    def __cinit__(Simulation self):
        self.data = <SimuData *>0
  
    def __dealloc__(Simulation self):
        if self.data != <SimuData *>0:
          del self.data


class PySimulationAdaptor(PySimulationBase):
  """
  PySimulationAdaptor(PySimulationBase_)
  
  This class is an adaptor for an internal type to the loader. It defines
  all the methods of PySimulationBase.
  
  Attributes:
    simu: a Simulation_ object
  """
  def __init__(self,sim):
    self.simu = sim

  def getBoxsize(self):
    return self.simu.BoxSize
    
  def getPositions(self):
    return self.simu.positions

  def getTypes(self):
    return self.simu.types
    
  def getVelocities(self):
    return self.simu.velocities
    
  def getIdentifiers(self):
    return self.simu.identifiers

  def getTime(self):
    return self.simu.time
    
  def getHubble(self):
    return self.simu.Hubble
    
  def getOmega_M(self):
    return self.simu.Omega_M
    
  def getOmega_Lambda(self):
    return self.simu.Omega_Lambda

  def getMasses(self):
    return self.simu.masses

cdef class ArrayWrapper:
    cdef void* data_ptr
    cdef np.uint64_t size
    cdef int type_array
    
    cdef set_data(self, np.uint64_t size, int type_array, void* data_ptr):
        """ Set the data of the array
            This cannot be done in the constructor as it must recieve C-level
            arguments.

            Args:
                size (int): Length of the array.
                data_ptr (void*): Pointer to the data
        """
        self.data_ptr = data_ptr
        self.size = size
        self.type_array = type_array
     
    def __array__(self):
        """ Here we use the __array__ method, that is called when numpy
tries to get an array from the object."""
        cdef np.npy_intp shape[1]
        
        shape[0] = <np.npy_intp> self.size
        # Create a 1D array, of length 'size'
        ndarray = np.PyArray_SimpleNewFromData(1, shape, self.type_array, self.data_ptr)
        return ndarray
     
    def __dealloc__(self):
        """ Frees the array. This is called by Python when all the
references to the object are gone. """
        pass
 
cdef object wrap_array(void *p, np.uint64_t s, int typ):
    cdef np.ndarray ndarray
    cdef ArrayWrapper wrapper

    wrapper = ArrayWrapper()
    wrapper.set_data(s, typ, p)
    ndarray = np.array(wrapper, copy=False)
    ndarray.base = <PyObject*> wrapper
    Py_INCREF(wrapper)
    
    return ndarray


cdef object wrap_float_array(float *p, np.uint64_t s):
    return wrap_array(<void *>p, s, np.NPY_FLOAT32)

cdef object wrap_int64_array(int64_t* p, np.uint64_t s):
    return wrap_array(<void *>p, s, np.NPY_INT64)

cdef object wrap_int_array(int* p, np.uint64_t s):
    return wrap_array(<void *>p, s, np.NPY_INT)

cdef object wrap_simudata(SimuData *data, int flags):
  cdef Simulation simu

  simu = Simulation()
  simu.data = data
  if flags & NEED_POSITION:
     simu.positions = [wrap_float_array(data.Pos[i], data.NumPart) for i in xrange(3)]
  else:
     simu.positions = None

  if flags & NEED_VELOCITY:
     simu.velocities = [wrap_float_array(data.Vel[i], data.NumPart) for i in xrange(3)]
  else:
     simu.velocities = None

  if flags & NEED_GADGET_ID:
     simu.identifiers = wrap_int64_array(data.Id, data.NumPart)
  else:
     simu.identifiers = None

  if flags & NEED_TYPE:
     simu.types = wrap_int_array(data.type, data.NumPart)
  else:
     simu.types = None

  if flags & NEED_MASS:
     simu.masses = wrap_float_array(data.Mass, data.NumPart)
  else:
     simu.masses = None

  return simu

def loadGadget(str filename, int snapshot_id, int gadgetFormat = 1, bool loadPosition = True, bool loadVelocity = False, bool loadId = False, bool loadType = False, bool loadMass=False):
    """loadGadget(filename, snapshot_id, gadgetFormat = 1, loadPosition=True, loadVelocity=False, loadId=False, loadType=False)

    This function loads Gadget-1 snapshot format.
    
    If snapshot_id is negative then the snapshot is considered not to be part of
    a set of snapshots written by different cpu. Otherwise the filename is modified
    to reflect the indicated snapshot_id.

    Arguments:
        filename (str): input filename
        snapshot_id (int): identifier of the gadget file if it is a multi-file snapshot
    
    Keyword arguments:
        loadPosition (bool): whether to load positions
        loadVelocity (bool): whether to load velocities
        loadId (bool): whether to load unique identifiers
        loadType (bool): whether to set types to particles
        loadMass (bool): whether to set the mass of particles
    
    Returns:
        an PySimulationAdaptor instance. 
"""

    cdef int flags
    cdef SimuData *data
    cdef Simulation simu
    cdef const char *filename_bs

    flags = 0
    if loadPosition:
        flags |= NEED_POSITION
    if loadVelocity:
        flags |= NEED_VELOCITY
    if loadId:
        flags |= NEED_GADGET_ID
    if loadType:
        flags |= NEED_TYPE
    if loadMass:
        flags |= NEED_MASS

    filename_b = bytes(filename, 'utf-8')
    filename_bs = filename_b
    with nogil:
      data = loadGadgetMulti(filename_bs, snapshot_id, flags, gadgetFormat)
    if data == <SimuData*>0:
        return None

    return PySimulationAdaptor(wrap_simudata(data, flags))

def loadParallelGadget(object filename_list, int gadgetFormat = 1, bool loadPosition = True, bool loadVelocity = False, bool loadId = False, bool loadType = False, bool loadMass=False):
  """loadParallelGadget(filename list, gadgetFormat=1, loadPosition=True, loadVelocity=False, loadId=False, loadType=False)

  Arguments:
    filename (list): a list or tuple of filenames to load in parallel
  
  Keyword arguments:
    loadPosition (bool): indicate to load positions
    loadVelocity (bool): indicate to load velocities
    loadId (bool): indicate to load id
    loadType (bool): indicate to load particle types
    
  Returns:
    It loads a gadget-1 snapshot and return a cosmotool.PySimulationBase_ object.
  """
  cdef int flags, i, num_files
  cdef list out_arrays
  cdef SimuData ** data
  cdef SimuData * local_data
  cdef Simulation simu
  cdef cppvector[cppstring.string] filenames

  flags = 0
  if loadPosition:
     flags |= NEED_POSITION
  if loadVelocity:
     flags |= NEED_VELOCITY
  if loadId:
     flags |= NEED_GADGET_ID
  if loadType:
     flags |= NEED_TYPE
  if loadMass:
     flags |= NEED_MASS

  num_files = len(filename_list)
  filenames.resize(num_files)
  data = alloc_simudata(num_files)
  for i,l in enumerate(filename_list):
    filenames[i] = l.encode('utf-8')
  
  with nogil:
    for i in prange(num_files):
      local_data = loadGadgetMulti_safe(filenames[i], flags, gadgetFormat)
      data[i] = local_data
#      data[i] = loadGadgetMulti(filenames[i].c_str(), -1, flags)
  
  out_arrays = []
  for i in xrange(num_files):
    if data[i] == <SimuData*>0:        
      out_arrays.append(None)
    else:
      out_arrays.append(PySimulationAdaptor(wrap_simudata(data[i], flags)))

  del_simudata(data)

  return out_arrays

def writeGadget(str filename, object simulation):
  """writeGadget(filename, simulation)
  
  This function attempts to write the content of the simulation object into 
  a file named `filename` using a Gadget-1 file format.
  
  Arguments:
    filename (str): output filename
    simulation (PySimulationBase): a simulation object
  """
  cdef SimuData simdata
  cdef np.ndarray[np.float32_t, ndim=1] pos, vel
  cdef np.ndarray[np.int64_t, ndim=1] ids
  cdef np.int64_t NumPart
  cdef int j
  
  if not isinstance(simulation,PySimulationBase):
    raise TypeError("Second argument must be of type SimulationBase")
  
  NumPart = simulation.positions[0].size
  simdata.noAuto = True
  
  for j in xrange(3):
    pos = simulation.getPositions()[j]
    vel = simulation.getVelocities()[j]
    
    if pos.size != NumPart or vel.size != NumPart:
      raise ValueError("Invalid number of particles")
    
    simdata.Pos[j] = <float *>pos.data
    simdata.Vel[j] = <float *>vel.data
  
  ids = simulation.getIdentifiers()
  simdata.Id = <int64_t *>ids.data
  simdata.BoxSize = simulation.getBoxsize()
  simdata.time = simulation.getTime()
  simdata.Hubble = simulation.getHubble()
  simdata.Omega_M = simulation.getOmega_M()
  simdata.Omega_Lambda = simulation.getOmega_Lambda()
  simdata.TotalNumPart = NumPart
  simdata.NumPart = NumPart

  filename_b = bytes(filename, 'utf-8')
  cxx_writeGadget(filename_b, &simdata)

def loadRamses(str basepath, int snapshot_id, int cpu_id, bool doublePrecision = False, bool loadPosition = True, bool loadVelocity = False, bool loadId = False, bool loadMass = False):
  """ loadRamses(basepath, snapshot_id, cpu_id, doublePrecision = False, loadPosition = True, loadVelocity = False)
  Loads the indicated snapshot based on the cpu id, snapshot id and basepath. It is important to specify the correct precision in doublePrecision otherwise the loading will fail. There is no way of auto-detecting properly the precision of the snapshot file.
  
  Args:
    basepath (str): the base directory of the snapshot
    snapshot_id (int): the snapshot id
    cpu_id (int): the cpu id of the file to load
  
  Keyword args:
    doublePrecision (bool): By default it is False, thus singlePrecision
    loadPosition (bool): Whether to load positions
    loadVelocity (bool): Whether to load velocities
    loadId (bool): Whether to load identifiers
    loadMass (bool): Whether to load mass value 
    
  Returns:
    An object derived from PySimulationBase_. 
"""
  cdef int flags
  cdef SimuData *data
  cdef Simulation simu

  flags = 0
  if loadPosition:
     flags |= NEED_POSITION
  if loadVelocity:
     flags |= NEED_VELOCITY
  if loadId:
     flags |= NEED_GADGET_ID
  if loadMass:
     flags |= NEED_MASS

  encpath = basepath.encode('utf-8') 
  try:
    data = loadRamsesSimu(encpath, snapshot_id, cpu_id, doublePrecision, flags)
    if data == <SimuData*>0:
      return None
  except RuntimeError as e:
    raise RuntimeError(str(e) + ' (check the float precision in snapshot)')

  return PySimulationAdaptor(wrap_simudata(data, flags))
  



from libcpp cimport bool
from libcpp cimport string as cppstring
from libcpp cimport vector as cppvector
import numpy as np
cimport numpy as np
from cpython cimport PyObject, Py_INCREF
cimport cython

np.import_array()

cdef extern from "cic.hpp" namespace "CosmoTool":

  ctypedef float CICType
  ctypedef float Coordinates[3]
  
  cdef cppclass CICParticles:
    float mass
    Coordinates coords

  cdef cppclass CICFilter:
  
    CICFilter(np.uint32_t resolution, double L) nogil
    
    void resetMesh() nogil
    void putParticles(CICParticles* particles, np.uint32_t N) nogil
    void getDensityField(CICType *& field, np.uint32_t& res) nogil


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def leanCic(float[:,:] particles, float L, int Resolution):
  cdef cppvector.vector[CICParticles] *p
  cdef CICFilter *cic
  cdef np.uint64_t i
  cdef CICType *field
  cdef np.uint32_t dummyRes
  cdef np.ndarray[np.float64_t, ndim=3] out_field
  cdef np.ndarray[np.float64_t, ndim=1] out_field0
  cdef np.float64_t[:] out_field_buf
  cdef np.uint64_t j

  cic = new CICFilter(Resolution, L)
  print("Reset mesh")
  cic.resetMesh()
  
  if particles.shape[1] != 3:
    raise ValueError("Particles must be Nx3 array")
  
  print("Inserting particles")
#  p = new cppvector.vector[CICParticles](particles.shape[0])
#  for i in xrange(particles.shape[0]):
#    *p[i].mass = 1
#    *p[i].coords[0] = particles[i,0]
#    *p[i].coords[1] = particles[i,1]
#    *p[i].coords[2] = particles[i,2]

#  cic.putParticles(&p[0], particles.shape[0])
  del p
    
  print("Done")
  field = <CICType*>0
  dummyRes = 0
  cic.getDensityField(field, dummyRes)

  print("Got to allocate a numpy %dx%dx%d" % (dummyRes, dummyRes,dummyRes))
  
  out_field = np.empty((dummyRes, dummyRes, dummyRes), dtype=np.float64)
  out_field0 = out_field.reshape(out_field.size)
  out_field_buf = out_field
  print("Copy")
  for j in xrange(out_field_buf.size):
    out_field_buf[j] = field[j]
  
  del cic
  return out_field



import numexpr as ne
import numpy as np
import cosmotool as ct
import icgen as bic
import icgen.cosmogrowth as cg
import sys
import argparse

cosmo={'omega_M_0':0.3175, 'h':0.6711}
cosmo['omega_lambda_0']=1-cosmo['omega_M_0']
cosmo['omega_k_0'] = 0
cosmo['omega_B_0']=0.049
cosmo['SIGMA8']=0.8344
cosmo['ns']=0.9624
N0=256

doSimulation=True

astart=1.

parser=argparse.ArgumentParser(description="Generate CIC density from 2LPT")
parser.add_argument('--start', type=int, required=True)
parser.add_argument('--end', type=int, required=True)
parser.add_argument('--step', type=int, required=True)
parser.add_argument('--base', type=str, required=True)
parser.add_argument('--N', type=int, default=256)
parser.add_argument('--output', type=str, default="dcic_%d.npy")
parser.add_argument('--supersample', type=int, default=1)
parser.add_argument('--rsd', action='store_true') 
args = parser.parse_args()



for i in xrange(args.start, args.end, args.step):
  print i
#  pos,_,density,N,L,_ = bic.run_generation("/nethome/lavaux/remote/borg_2m++_128/initial_density_%d.dat" % i, 0.001, astart, cosmo, supersample=2, do_lpt2=True)
  pos,vel,density,N,L,_,_ = bic.run_generation("%s/initial_density_%d.dat" % (args.base,i), 0.001, astart, 
                                           cosmo, supersample=args.supersample, do_lpt2=True, needvel=True)

  if args.rsd:
    inv_r2 = ne.evaluate('1/sqrt(x**2+y**2+z**2)', 
        local_dict={'x':pos[0], 'y':pos[1], 'z':pos[2]})
    rsd = lambda p,v: ne.evaluate('x + (x*vx)*inv_r2 / H', 
        local_dict={'x':p, 'inv_r2':inv_r2,
         'vx':v, 'H':100.0}, out=p, casting='unsafe')

    rsd(pos[0], vel[0])
    rsd(pos[1], vel[1])
    rsd(pos[2], vel[2])

  dcic = ct.cicParticles(pos, L, args.N)
  dcic /= np.average(np.average(np.average(dcic, axis=0), axis=0), axis=0)
  dcic -= 1

  np.save(args.output % i, dcic)



cimport cython
cimport numpy as npx

ctypedef fused sum_type:
    cython.int
    cython.float
    npx.uint64_t
    npx.uint32_t

@cython.boundscheck(False)
cdef inline sum_type _mysum(sum_type[:] jobs) nogil:
  cdef sum_type s
  cdef npx.uint64_t N
  cdef int i

  s = 0
  N = jobs.shape[0]
  for i in xrange(N):
    s += jobs[i]
  return s





import healpy as hp
import numpy as np
import cosmotool as ct
import argparse
import h5py as h5
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import ksz
from ksz.constants import *
from cosmotool import interp3d

def wrapper_impulsion(f):

  class _Wrapper(object):
    def __init__(self):
      pass
      
    def __getitem__(self,direction):
    
      if 'velocity' in f:
        return f['velocity'][:,:,:,direction]
      
      n = "p%d" % direction
      return f[n]

  return _Wrapper()

parser=argparse.ArgumentParser(description="Generate Skymaps from CIC maps")
parser.add_argument('--boxsize', type=float, required=True)
parser.add_argument('--Nside', type=int, default=128)
parser.add_argument('--base_h5', type=str, required=True)
parser.add_argument('--base_fig', type=str, required=True)
parser.add_argument('--start', type=int, required=True)
parser.add_argument('--end', type=int, required=True)
parser.add_argument('--step', type=int, required=True)
parser.add_argument('--minval', type=float, default=-0.5)
parser.add_argument('--maxval', type=float, default=0.5)
parser.add_argument('--depth_min', type=float, default=10)
parser.add_argument('--depth_max', type=float, default=60)
parser.add_argument('--iid', type=int, default=0)
parser.add_argument('--ksz_map', type=str, required=True)
args = parser.parse_args()

L = args.boxsize
Nside = args.Nside

def build_unit_vectors(N):
  ii = np.arange(N)*L/N - 0.5*L
  d = np.sqrt(ii[:,None,None]**2 + ii[None,:,None]**2 + ii[None,None,:]**2)
  d[N/2,N/2,N/2] = 1
  ux = ii[:,None,None] / d
  uy = ii[None,:,None] / d
  uz = ii[None,None,:] / d

  return ux,uy,uz

def build_radial_v(v):
  N = v[0].shape[0]
  u = build_unit_vectors(N)
  vr = v[0] * u[2]
  vr += v[1] * u[1]
  vr += v[2] * u[0]

  return vr.transpose()

def generate_from_catalog(vfield,Boxsize,dmin,dmax):
  import progressbar as pbar
  
  cat = np.load("2m++.npy")


  cat['distance'] = cat['best_velcmb']
  cat = cat[np.where((cat['distance']>100*dmin)*(cat['distance']<dmax*100))]

  deg2rad = np.pi/180
  Npix = 12*Nside**2
  xp,yp,zp = hp.pix2vec(Nside, np.arange(Npix))
  N2 = np.sqrt(xp**2+yp**2+zp**2)

  ksz_template = np.zeros(Npix, dtype=np.float64)
  ksz_mask = np.zeros(Npix, dtype=np.uint8)

  pb = pbar.ProgressBar(maxval = cat.size, widgets=[pbar.Bar(), pbar.ETA()]).start()

  for k,i in np.ndenumerate(cat):
    pb.update(k[0])
    l,b=i['gal_long'],i['gal_lat']
    ra,dec=i['ra'],i['dec']

    l *= deg2rad
    b *= deg2rad

    ra *= deg2rad
    dec *= deg2rad

#    x0 = np.cos(l)*np.cos(b)
#    y0 = np.sin(l)*np.cos(b)
#    z0 = np.sin(b)

    x0 =  xra = np.cos(ra)*np.cos(dec)
    y0 = yra = np.sin(ra)*np.cos(dec)
    z0 = zra = np.sin(dec)
    
    DA =i['distance']/100
    Lgal = DA**2*10**(0.4*(tmpp_cat['Msun']-i['K2MRS']+25))

    profiler = ksz.KSZ_Isothermal(Lgal, 2.37)

    idx0 = hp.query_disc(Nside, (x0,y0,z0), 3*profiler.rGalaxy/DA)

    vr = interp3d(DA * xra, DA * yra, DA * zra, vfield, Boxsize)

    xp1 = xp[idx0]
    yp1 = yp[idx0]
    zp1 = zp[idx0]
    N2_1 = N2[idx0]

    cos_theta = x0*xp1+y0*yp1+z0*zp1
    cos_theta /= np.sqrt(x0**2+y0**2+z0**2)*(N2_1)

    idx,idx_masked,m = profiler.projected_profile(cos_theta, DA)
    idx = idx0[idx]
    idx_masked = idx0[idx_masked]
    ksz_template[idx] += vr * m
    ksz_mask[idx_masked] = 0

  pb.finish()

  return ksz_template, ksz_mask

for i in xrange(args.start,args.end,args.step):
  ff=plt.figure(1)
  plt.clf()
  v=[]
  fname = args.base_h5 % i
  if False:
    print("Opening %s..." % fname)
    with h5.File(fname, mode="r") as f:
      p = wrapper_impulsion(f)
      for j in xrange(3):
        v.append(p[j] / f['density'][:])

    print("Building radial velocities...")

  # Rescale by Omega_b / Omega_m
    vr = build_radial_v(v)
#  _,_,vr = build_unit_vectors(128)
#  vr *= 1000 * 500
    vr *= ksz_normalization*1e6
    del v
#    np.save("vr.npy", vr)
  else:
    print("Loading vrs...")
    vr = np.load("vr.npy")

  print("Generating map...")

  proj,mask = generate_from_catalog(vr,args.boxsize,args.depth_min,args.depth_max)

  hp.write_map(args.ksz_map % i, proj)
  hp.write_map((args.ksz_map % i) + "_mask", mask)

  hp.mollview(proj, fig=1, coord='CG', cmap=plt.cm.coolwarm, title='Sample %d' % i, min=args.minval, 
              max=args.maxval)

  ff.savefig(args.base_fig % i)



from libcpp cimport bool
from libcpp cimport string as cppstring
import numpy as np
cimport numpy as np
from cpython cimport PyObject, Py_INCREF
cimport cython

np.import_array()

cdef extern from "cosmopower.hpp" namespace "CosmoTool":

  cdef enum CosmoFunction "CosmoTool::CosmoPower::CosmoFunction":
      POWER_EFSTATHIOU "CosmoTool::CosmoPower::POWER_EFSTATHIOU",
      HU_WIGGLES "CosmoTool::CosmoPower::HU_WIGGLES",
      HU_BARYON "CosmoTool::CosmoPower::HU_BARYON",
      OLD_POWERSPECTRUM,
      POWER_BARDEEN "CosmoTool::CosmoPower::POWER_BARDEEN",
      POWER_SUGIYAMA "CosmoTool::CosmoPower::POWER_SUGIYAMA",
      POWER_BDM,
      POWER_TEST,
      HU_WIGGLES_ORIGINAL "CosmoTool::CosmoPower::HU_WIGGLES_ORIGINAL"

  cdef cppclass CosmoPower:
    double n
    double K0
    double V_LG_CMB
    
    double CMB_VECTOR[3]
    double h
    double SIGMA8
    double OMEGA_B
    double OMEGA_C
    double omega_B
    double omega_C
    double Theta_27
    double OMEGA_0
    double Omega
    double beta
    double OmegaEff
    double Gamma0
    double normPower
    
        
    CosmoPower()
    void setFunction(CosmoFunction)
    void updateCosmology()
    void updatePhysicalCosmology()
    void normalize(double,double)
    void setNormalization(double)
    double power(double)
 
cdef class CosmologyPower:
  """CosmologyPower(**cosmo)
  
  CosmologyPower manages and compute power spectra computation according to different
  approximation given in the litterature.
  
  Keyword arguments:
    omega_B_0 (float): relative baryon density
    omega_M_0 (float): relative matter density
    h (float): Hubble constant relative to 100 km/s/Mpc
    ns (float): power law of the large scale inflation spectrum
  """

  cdef CosmoPower power

  def __init__(self,**cosmo):    
    """Constructor

    Keyword arguments:
      * omega_B_0
      * omega_M_0
      * h
      * ns
      * T27
    """
    self.power = CosmoPower()
    self.power.OMEGA_B = cosmo['omega_B_0']
    self.power.OMEGA_C = cosmo['omega_M_0']-cosmo['omega_B_0']
    self.power.h = cosmo['h']
    if 'ns' in cosmo:
      self.power.n = cosmo['ns'] 
    if 'T27' in cosmo:
      self.power.Theta_27 = cosmo['T27']
    
    assert self.power.OMEGA_C > 0
    
    self.power.updateCosmology()

  def setNormalization(self,A):
    """Set manual normalization for A_S"""
    self.power.setNormalization(A)
    
  def normalize(self,s8,k_min=-1,k_max=-1):
    """normalize(self, sigma8)
    
    Compute the normalization of the power spectrum using sigma8.
    
    Arguments:
      sigma8 (float): standard deviation of density field smoothed at 8 Mpc/h
    """
    self.power.SIGMA8 = s8
    self.power.normalize(k_min, k_max)
    
    
  def setFunction(self,funcname):
    """setFunction(self, funcname)
    
    Choose an approximation to use for the computation of the power spectrum
    
    Arguments:
      funcname (str): the name of the approximation. It can be either
                      EFSTATHIOU, HU_WIGGLES, HU_BARYON, BARDEEN or SUGIYAMA.
    """
    cdef CosmoFunction f
    
    f = POWER_EFSTATHIOU
  
    if funcname=='EFSTATHIOU':
      f = POWER_EFSTATHIOU
    elif funcname=='HU_WIGGLES':
      f = HU_WIGGLES
    elif funcname=='HU_BARYON':
      f = HU_BARYON
    elif funcname=='BARDEEN':
      f = POWER_BARDEEN
    elif funcname=='SUGIYAMA':
      f = POWER_SUGIYAMA
    elif funcname=='HU_WIGGLES_ORIGINAL':
      f = HU_WIGGLES_ORIGINAL
    else:
      raise ValueError("Unknown function name " + funcname)

    self.power.setFunction(f)
    
  cdef double _compute(self, double k):
    k *= self.power.h
    return self.power.power(k) * self.power.h**3
    
  def compute(self, k):
    """compute(self, k)
    
    Compute the power spectrum for mode which length k.
    
    Arguments:
      k (float): Mode for which to evaluate the power spectrum.
                 It can be a scalar or a numpy array. 
                 The units must be in 'h Mpc^{-1}'.
    
    Returns:
      a scalar or a numpy array depending on the type of the k argument 
    """
    
    cdef np.ndarray out
    cdef double kval
    cdef tuple i
    
    if isinstance(k, np.ndarray):
      out = np.empty(k.shape, dtype=np.float64)
      for i,kval in np.ndenumerate(k):
        out[i] = self._compute(kval)
      return out
    else:
      return self._compute(k)
      



import cosmotool as ct
import numpy as np
import cosmolopy as cpy
from cosmogrowth import *
import borgadaptor as ba

@ct.timeit
def gen_posgrid(N, L, delta=1, dtype=np.float32):
    """ Generate an ordered lagrangian grid"""

    ix = (np.arange(N)*(L/N*delta)).astype(dtype)

    x = ix[:,None,None].repeat(N, axis=1).repeat(N, axis=2)
    y = ix[None,:,None].repeat(N, axis=0).repeat(N, axis=2)
    z = ix[None,None,:].repeat(N, axis=0).repeat(N, axis=1)

    return x.reshape((x.size,)), y.reshape((y.size,)), z.reshape((z.size,))

def bin_power(P, L, bins=20, range=(0,1.), dev=False):

  N = P.shape[0]
  ik = np.fft.fftfreq(N, d=L/N)*2*np.pi

  k = np.sqrt(ik[:,None,None]**2 + ik[None,:,None]**2 + ik[None,None,:(N/2+1)]**2)

  H,b = np.histogram(k, bins=bins, range=range)
  Hw,b = np.histogram(k, bins=bins, weights=P, range=range)
  
  if dev:
    return Hw/(H-1), 0.5*(b[1:]+b[0:bins]), 1.0/np.sqrt(H)
  else:
    return Hw/(H-1), 0.5*(b[1:]+b[0:bins])

def compute_power_from_borg(input_borg, a_borg, cosmo, bins=10, range=(0,1)):
  borg_vol = ct.read_borg_vol(input_borg)
  N = borg_vol.density.shape[0]

  cgrowth = CosmoGrowth(**cosmo)
  D1 = cgrowth.D(1)
  D1_0 = D1/cgrowth.D(a_borg)
  print("D1_0=%lg" % D1_0)

  density_hat, L = ba.half_pixel_shift(borg_vol)

  return bin_power(D1_0**2*np.abs(density_hat)**2/L**3, L, bins=bins, range=range)
  
def compute_ref_power(L, N, cosmo, bins=10, range=(0,1), func='HU_WIGGLES'):
  ik = np.fft.fftfreq(N, d=L/N)*2*np.pi

  k = np.sqrt(ik[:,None,None]**2 + ik[None,:,None]**2 + ik[None,None,:(N/2+1)]**2)
  p = ct.CosmologyPower(**cosmo)
  p.setFunction(func)
  p.normalize(cosmo['SIGMA8'])

  return bin_power(p.compute(k)*cosmo['h']**3, L, bins=bins, range=range)


def do_supergenerate(density, density_out=None, mulfac=None,zero_fill=False,Pk=None,L=None,h=None):
 
  N = density.shape[0]
  
  if density_out is None:
    assert mulfac is not None
    Ns = mulfac*N
    density_out = np.zeros((Ns,Ns,Ns/2+1), dtype=np.complex128)
    density_out[:] = np.nan
  elif mulfac is None:
    mulfac = density_out.shape[0] / N
    Ns = density_out.shape[0]
    assert (density_out.shape[0] % N) == 0
  	
  assert len(density_out.shape) == 3
  assert density_out.shape[0] == density_out.shape[1]
  assert density_out.shape[2] == (density_out.shape[0]/2+1)
  
  hN = N/2
  density_out[:hN,               :hN, :hN+1] = density[:hN, :hN, :]
  density_out[:hN,        (Ns-hN):Ns, :hN+1] = density[:hN, hN:, :]
  density_out[(Ns-hN):Ns, (Ns-hN):Ns, :hN+1] = density[hN:, hN:, :]
  density_out[(Ns-hN):Ns,        :hN, :hN+1] = density[hN:, :hN, :]

  if mulfac > 1:

    cond=np.isnan(density_out)
    if zero_fill:
      density_out[cond] = 0
    else:

      if Pk is not None:
        assert L is not None and h is not None

      @ct.timeit_quiet
      def build_Pk():
         ik = np.fft.fftfreq(Ns, d=L/Ns)*2*np.pi
         k = ne.evaluate('sqrt(kx**2 + ky**2 + kz**2)', {'kx':ik[:,None,None], 'ky':ik[None,:,None], 'kz':ik[None,None,:(Ns/2+1)]})
         return Pk.compute(k)*L**3

      print np.where(np.isnan(density_out))[0].size
      Nz = np.count_nonzero(cond)
      amplitude = np.sqrt(build_Pk()[cond]/2) if Pk is not None else (1.0/np.sqrt(2))
      density_out.real[cond] = np.random.randn(Nz) * amplitude
      density_out.imag[cond] = np.random.randn(Nz) * amplitude
      print np.where(np.isnan(density_out))[0].size
    
      # Now we have to fix the Nyquist plane
      hNs = Ns/2
      nyquist = density_out[:, :, hNs]
      Nplane = nyquist.size
      nyquist.flat[:Nplane/2] = np.sqrt(2.0)*nyquist.flat[Nplane:Nplane/2:-1].conj()


  return density_out

@ct.timeit_quiet
def run_generation(input_borg, a_borg, a_ic, cosmo, supersample=1, supergenerate=1, do_lpt2=True, shiftPixel=False, psi_instead=False, needvel=True, func='HU_WIGGLES'):
    """ Generate particles and velocities from a BORG snapshot. Returns a tuple of 
    (positions,velocities,N,BoxSize,scale_factor)."""

    borg_vol = ct.read_borg_vol(input_borg)
    N = borg_vol.density.shape[0]

    cgrowth = CosmoGrowth(**cosmo)

    density, L = ba.half_pixel_shift(borg_vol, doshift=shiftPixel)


    # Compute LPT scaling coefficient
    D1 = cgrowth.D(a_ic)
    D1_0 = D1/cgrowth.D(a_borg)
    Dborg = cgrowth.D(a_borg)/cgrowth.D(1.0)
    print "D1_0=%lg" % D1_0

    if supergenerate>1:
      print("Doing supergeneration (factor=%d)" % supergenerate)
      p = ct.CosmologyPower(**cosmo)
      p.setFunction(func)
      p.normalize(cosmo['SIGMA8']*Dborg)
      density = do_supergenerate(density,mulfac=supergenerate,Pk=p,L=L,h=cosmo['h'])

    lpt = LagrangianPerturbation(-density, L, fourier=True, supersample=supersample)

    # Generate grid
    posq = gen_posgrid(N*supersample, L)
    vel= []
    posx = []
    
    velmul = cgrowth.compute_velmul(a_ic) if not psi_instead else 1
    
    D2 = -3./7 * D1_0**2

    if do_lpt2:
      psi2 = lpt.lpt2('all')
    for j in xrange(3):
        # Generate psi_j (displacement along j)
        print("LPT1 axis=%d" % j)
        psi = D1_0*lpt.lpt1(j)
        psi = psi.reshape((psi.size,))
        if do_lpt2:
          print("LPT2 axis=%d" % j)
          psi += D2 * psi2[j].reshape((psi2[j].size,))
        # Generate posx
        posx.append(((posq[j] + psi)%L).astype(np.float32))
        # Generate vel
        if needvel:
          vel.append((psi*velmul).astype(np.float32)) 

    print("velmul=%lg" % (cosmo['h']*velmul))

    lpt.cube.dhat = lpt.dhat
    density = lpt.cube.irfft()
    density *= (cgrowth.D(1)/cgrowth.D(a_borg))

    return posx,vel,density,N*supergenerate*supersample,L,a_ic,cosmo


@ct.timeit_quiet
def whitify(density, L, cosmo, supergenerate=1, zero_fill=False, func='HU_WIGGLES'):

  N = density.shape[0]
  p = ct.CosmologyPower(**cosmo)
  p.setFunction(func)
  p.normalize(cosmo['SIGMA8'])

  @ct.timeit_quiet
  def build_Pk():
    ik = np.fft.fftfreq(N, d=L/N)*2*np.pi
    k = np.sqrt(ik[:,None,None]**2 + ik[None,:,None]**2 + ik[None,None,:(N/2+1)]**2)
    return p.compute(k)*L**3
    
  Pk = build_Pk()
  Pk[0,0,0]=1

  cube = CubeFT(L, N)
  cube.density = density
  density_hat = cube.rfft()
  density_hat /= np.sqrt(Pk)

  Ns = N*supergenerate

  density_hat_super = do_supergenerate(density_hat, mulfac=supergenerate)

  cube = CubeFT(L, Ns)
  cube.dhat = density_hat_super
  return np.fft.irfftn(density_hat_super)*Ns**1.5


    
def write_icfiles(*generated_ic, **kwargs):
  """Write the initial conditions from the tuple returned by run_generation"""
  
  supergenerate=kwargs.get('supergenerate', 1)
  zero_fill=kwargs.get('zero_fill', False)
  posx,vel,density,N,L,a_ic,cosmo = generated_ic

  ct.simpleWriteGadget("Data/borg.gad", posx, velocities=vel, boxsize=L, Hubble=cosmo['h'], Omega_M=cosmo['omega_M_0'], time=a_ic) 
  for i,c in enumerate(["z","y","x"]):
    ct.writeGrafic("Data/ic_velc%s" % c, vel[i].reshape((N,N,N)), L, a_ic, **cosmo)

  ct.writeGrafic("Data/ic_deltab", density, L, a_ic, **cosmo)
  
  ct.writeWhitePhase("Data/white.dat", whitify(density, L, cosmo, supergenerate=supergenerate,zero_fill=zero_fill))
  
  with file("Data/white_params", mode="w") as f:
    f.write("4\n%lg, %lg, %lg\n" % (cosmo['omega_M_0'], cosmo['omega_lambda_0'], 100*cosmo['h']))
    f.write("%lg\n%lg\n-%lg\n0,0\n" % (cosmo['omega_B_0'],cosmo['ns'],cosmo['SIGMA8']))
    f.write("-%lg\n1\n0\n\n\n\n\n" % L)
    f.write("2\n\n0\nwhite.dat\n0\npadding_white.dat\n")




import struct
import numpy as np

def readGrafic(filename):
  """This function reads a grafic file.
  
  Arguments:
    filename (str): the path to the grafic file
    
  Returns:
    a tuple containing:
      * the array held in the grafic file
      * the size of the box
      * the scale factor
      * the mean matter density :math:`\Omega_\mathrm{m}`
      * the dark energy density :math:`\Omega_\Lambda`
      * the hubble constant, relative to 100 km/s/Mpc
      * xoffset
      * yoffset
      * zoffset
  """ 
  with open(filename, mode="rb") as f:
        p = struct.unpack("IIIIffffffffI", f.read(4*11 + 2*4))
        checkPoint0, Nx, Ny, Nz, delta, xoff, yoff, zoff, scalefac, omega0, omegalambda0, h, checkPoint1 = p
        if checkPoint0 != checkPoint1 or checkPoint0 != 4*11:
          raise ValueError("Invalid unformatted access")
          
        a = np.empty((Nx,Ny,Nz), dtype=np.float32)
        
        BoxSize = delta * Nx * h
        xoff *= h
        yoff *= h
        zoff *= h
        
        checkPoint = 4*Ny*Nz
        for i in range(Nx):
            checkPoint = struct.unpack("I", f.read(4))[0]
            if checkPoint != 4*Ny*Nz:
              raise ValueError("Invalid unformatted access")
              
            a[i, :, :] = np.fromfile(f, dtype=np.float32, count=Ny*Nz).reshape((Ny, Nz))

            checkPoint = struct.unpack("I", f.read(4))[0]
            if checkPoint != 4*Ny*Nz:
              raise ValueError("Invalid unformatted access")

  return a, BoxSize, scalefac, omega0, omegalambda0, h, xoff, yoff,zoff

def writeGrafic(filename, field, BoxSize, scalefac, **cosmo):

    with open(filename, mode="wb") as f:
        checkPoint = 4*11
        
        Nx,Ny,Nz = field.shape
        delta = BoxSize/Nx/cosmo['h']
        bad = 0.0
        
        f.write(struct.pack("IIIIffffffffI", checkPoint, 
                                          Nx, Ny, Nz, 
                                          delta, 
                                          bad, bad, bad, 
                                          scalefac,
                                          cosmo['omega_M_0'], cosmo['omega_lambda_0'], 100*cosmo['h'], checkPoint))
        checkPoint = 4*Ny*Nz
        field = field.reshape(field.shape, order='F')
        for i in range(Nx):
            f.write(struct.pack("I", checkPoint))
            f.write(field[i].astype(np.float32).tostring())
            f.write(struct.pack("I", checkPoint))
        
        
def writeWhitePhase(filename, field):

    with open(filename, mode="wb") as f:
        Nx,Ny,Nz = field.shape
        
        checkPoint = 4*4
        f.write(struct.pack("IIIIII", checkPoint, Nx, Ny, Nz, 0, checkPoint))
        
        field = field.reshape(field.shape, order='F')
        checkPoint = struct.pack("I", 4*Ny*Nz)
        for i in range(Nx):
            f.write(checkPoint)
            f.write(field[i].astype(np.float32).tostring())
            f.write(checkPoint)


def readWhitePhase(filename):
    with open(filename, mode="rb") as f:
        _, Nx, Ny, Nz, _, _ = struct.unpack("IIIIII", f.read(4*4+2*4))
        
        a = np.empty((Nx,Ny,Nz), dtype=np.float32)
        
        checkPoint_ref = 4*Ny*Nz
        
        for i in range(Nx):
            if struct.unpack("I", f.read(4))[0] != checkPoint_ref:
              raise ValueError("Invalid unformatted access")
              
            b = np.fromfile(f, dtype=np.float32, count=Ny*Nz).reshape((Ny, Nz))
            if i==0:
              print(b)
            a[i, : ,:] = b
            
            if struct.unpack("I", f.read(4))[0] != checkPoint_ref:
              raise ValueError("Invalid unformatted access")

    return a



import numpy as np
import cosmotool as ct
import borgicgen as bic
import cosmogrowth as cg
import sys

cosmo={'omega_M_0':0.3175, 'h':0.6711}
cosmo['omega_lambda_0']=1-cosmo['omega_M_0']
cosmo['omega_k_0'] = 0
cosmo['omega_B_0']=0.049
cosmo['SIGMA8']=0.8344
cosmo['ns']=0.9624
N0=256

doSimulation=False
simShift=False

snap_id=int(sys.argv[1])
astart=1/100.

if doSimulation:
  s = ct.loadRamsesAll("/nethome/lavaux/remote2/borgsim3/", snap_id, doublePrecision=True)
  astart=s.getTime()
  L = s.getBoxsize()

  p = s.getPositions()
  Nsim = int( np.round( p[0].size**(1./3)) )
  print("Nsim = %d" % Nsim)

  if simShift:
    p = [(q-0.5*L/Nsim)%L for q in p]

  dsim = ct.cicParticles(p[::-1], L, N0)
  dsim /= np.average(np.average(np.average(dsim, axis=0), axis=0), axis=0)
  dsim -= 1

  dsim_hat = np.fft.rfftn(dsim)*(L/N0)**3
  Psim, bsim = bic.bin_power(np.abs(dsim_hat)**2/L**3, L, range=(0,1.), bins=150)

pos,_,density,N,L,_,_ = bic.run_generation("initial_density_1872.dat", 0.001, astart, cosmo, supersample=1, do_lpt2=False, supergenerate=2)

dcic = ct.cicParticles(pos, L, N0)
dcic /= np.average(np.average(np.average(dcic, axis=0), axis=0), axis=0)
dcic -= 1

dcic_hat = np.fft.rfftn(dcic)*(L/N0)**3
dens_hat = np.fft.rfftn(density)*(L/N0)**3

Pcic, bcic = bic.bin_power(np.abs(dcic_hat)**2/L**3, L, range=(0,4.), bins=150)
Pdens, bdens = bic.bin_power(np.abs(dens_hat)**2/L**3, L, range=(0,4.), bins=150)

cgrowth = cg.CosmoGrowth(**cosmo)
D1 = cgrowth.D(astart)
D1_0 = D1/cgrowth.D(1)#0.001)

Pref, bref = bic.compute_ref_power(L, N0, cosmo, range=(0,4.), bins=150)

Pcic /= D1_0**2

#borg_evolved = ct.read_borg_vol("final_density_1380.dat")
#dborg_hat = np.fft.rfftn(borg_evolved.density)*L**3/borg_evolved.density.size

#Pborg, bborg = bic.bin_power(np.abs(dborg_hat)**2/L**3, L, range=(0,1.),bins=150)



import timeit
import numpy as np
import cosmotool as ct

def myfun(N):
  f=0.001
  L = 1.0
  d=np.random.normal(size=(N,)*3)   * np.sqrt(float(N))**3 / L**3
  rho = d + f *(d*d - np.average(d*d))
  delta = (L/N)**3
  
  B = ct.bispectrum(rho * delta, 1, N, fourier=False)
  P = ct.powerspectrum(rho * delta, 1, N, fourier=False)
  PP = P[1]/P[0] / L**3 
  
  x = PP[:,None,None] * PP[None,:,None] + PP[:,None,None]*PP[None,None,:] + PP[None,:,None]*PP[None,None,:]
  
  BB = B[1]/B[0] / L**3

  y = BB/x

  np.savez("bispec_%d.npz" % N, x=x, y=y, d=d,B_nt=B[0], B_r=B[1], P_n=P[0], P=P[1], BB=BB, rho=rho, PP=PP);
  
  
#print( timeit.timeit('from __main__ import myfun; myfun(16)', number=1) )
#print( timeit.timeit('from __main__ import myfun; myfun(24)', number=1) )
print( timeit.timeit('from __main__ import myfun; myfun(32)', number=1) )
#print( timeit.timeit('from __main__ import myfun; myfun(64)', number=1) )



import healpy as hp
import numpy as np
import cosmotool as ct
import h5py as h5
from matplotlib import pyplot as plt

L=600.
Nside=128

INDATA="/nethome/lavaux/Copy/PlusSimulation/BORG/Input_Data/2m++.npy"
tmpp = np.load(INDATA)

def build_sky_proj(density, dmax=60.,dmin=0):

  N = density.shape[0]
  ix = (np.arange(N)-0.5)*L/N - 0.5 * L


  dist2 = (ix[:,None,None]**2 + ix[None,:,None]**2 + ix[None,None,:]**2)

  flux = density.transpose().astype(ct.DTYPE) # / dist2
  dmax=N*dmax/L
  dmin=N*dmin/L
  projsky1 = ct.spherical_projection(Nside, flux, dmin, dmax, integrator_id=1)
#  projsky0 = ct.spherical_projection(Nside, flux, 0, 52, integrator_id=0)

  return projsky1*L/N#,projsky0

l,b = tmpp['gal_long'],tmpp['gal_lat']

l = np.radians(l)
b = np.pi/2 - np.radians(b)

dcmb = tmpp['velcmb']/100.

idx = np.where((dcmb>10)*(dcmb<60))

plt.figure(1)
plt.clf()
if True:
  with h5.File("fields.h5", mode="r") as f:
    d = f["density"][:].transpose()
    d /= np.average(np.average(np.average(d,axis=0),axis=0),axis=0)
    proj = build_sky_proj(d, dmin=10,dmax=60.)
  proj0 = proj1 = proj
else:
  d = np.load("icgen/dcic0.npy")
  proj0 = build_sky_proj(1+d, dmin=10,dmax=60.)
  d = np.load("icgen/dcic1.npy")
  proj1 = build_sky_proj(1+d, dmin=10,dmax=60.)

hp.mollview(proj0, fig=1, coord='CG', max=60, cmap=plt.cm.coolwarm)
hp.projscatter(b[idx], l[idx], lw=0, color='g', s=5.0, alpha=0.8)

plt.figure(2)
plt.clf()
hp.mollview(proj1, fig=2, coord='CG', max=60, cmap=plt.cm.coolwarm)
hp.projscatter(b[idx], l[idx], lw=0, color='g', s=5.0, alpha=0.8)



import numpy as np

try:
    import cffi
    import os

    _ffi = cffi.FFI()
    _ffi.cdef("""

    void CosmoTool_compute_bispectrum(
        double *delta_hat, size_t Nx, size_t Ny, size_t Nz,
        size_t *Ntriangles,
        double* B, double delta_k, size_t Nk ) ;
    void CosmoTool_compute_powerspectrum(
        double *delta_hat, size_t Nx, size_t Ny, size_t Nz,
        size_t *Ncounts,
        double* P, double delta_k, size_t Nk );

    """);

    _pathlib = os.path.dirname(os.path.abspath(__file__))
    _lib = _ffi.dlopen(os.path.join(_pathlib,"_cosmo_bispectrum.so"))
except Exception as e:
    print(repr(e))
    raise RuntimeError("Failed to initialize _cosmo_bispectrum module")


def bispectrum(delta, delta_k, Nk, fourier=True):
    """bispectrum(delta, fourier=True)
    
    Args:
        * delta: a 3d density field, can be Fourier modes if fourier set to True
        
        
    Return:
        * A 3d array of the binned bispectrum
"""

    if len(delta.shape) != 3:
        raise ValueError("Invalid shape for delta")
    try:
        delta_k = float(delta_k)
        Nk = int(Nk)
    except:
        raise ValueError()

    if not fourier:
        delta = np.fft.rfftn(delta)        
    N1,N2,N3 = delta.shape
    rN3 = (N3-1)*2
    delta_hat_buf = np.empty((N1*N2*N3*2),dtype=np.double)
    delta_hat_buf[::2] = delta.real.ravel()
    delta_hat_buf[1::2] = delta.imag.ravel()
    
    size_size = _ffi.sizeof("size_t")
    if size_size == 4:
        triangle_buf = np.zeros((Nk,Nk,Nk),dtype=np.int32)
    elif size_size == 8:
        triangle_buf = np.zeros((Nk,Nk,Nk),dtype=np.int64)
    else:
        raise RuntimeError("Internal error, do not know how to map size_t")

    B_buf = np.zeros((Nk*Nk*Nk*2), dtype=np.double)
    
    _lib.CosmoTool_compute_bispectrum( \
        _ffi.cast("double *", delta_hat_buf.ctypes.data), \
        N1, N2, rN3, \
        _ffi.cast("size_t *", triangle_buf.ctypes.data), \
        _ffi.cast("double *", B_buf.ctypes.data), \
        delta_k,  \
        Nk)
    B_buf = B_buf.reshape((Nk,Nk,Nk,2))
    return triangle_buf, B_buf[...,0]+1j*B_buf[...,1]

def powerspectrum(delta, delta_k, Nk, fourier=True):
    """powerspectrum(delta, fourier=True)
    
    Args:
        * delta: a 3d density field, can be Fourier modes if fourier set to True
        
        
    Return:
        * A 3d array of the binned bispectrum
"""

    if len(delta.shape) != 3:
        raise ValueError("Invalid shape for delta")
    try:
        delta_k = float(delta_k)
        Nk = int(Nk)
    except:
        raise ValueError()

    if not fourier:
        delta = np.fft.rfftn(delta)        
    N1,N2,N3 = delta.shape
    delta_hat_buf = np.empty((N1*N2*N3*2),dtype=np.double)
    delta_hat_buf[::2] = delta.real.ravel()
    delta_hat_buf[1::2] = delta.imag.ravel()
    
    size_size = _ffi.sizeof("size_t")
    if size_size == 4:
        count_buf = np.zeros((Nk,),dtype=np.int32)
    elif size_size == 8:
        count_buf = np.zeros((Nk,),dtype=np.int64)
    else:
        raise RuntimeError("Internal error, do not know how to map size_t")

    B_buf = np.zeros((Nk,), dtype=np.double)
    
    _lib.CosmoTool_compute_powerspectrum( \
        _ffi.cast("double *", delta_hat_buf.ctypes.data), \
        N1, N2, N3, \
        _ffi.cast("size_t *", count_buf.ctypes.data), \
        _ffi.cast("double *", B_buf.ctypes.data), \
        delta_k,  \
        Nk)
    return count_buf, B_buf[...]

        
if __name__=="__main__":
    delta=np.zeros((16,16,16))
    delta[0,0,0]=1
    delta[3,2,1]=1
    b = powerspectrum(delta, 1, 16, fourier=False)
    a = bispectrum(delta, 1, 16, fourier=False)
    print(a[0].max())



import numpy as np
from contextlib import contextmanager

class ProgrammableParticleLoad(object):

  @staticmethod
  def main_script(source, particles, aname="default", aux=None):
    import vtk
    from vtk.util import numpy_support as ns
  
    out = source.GetOutput()
    vv = vtk.vtkPoints()

    assert len(particles.shape) == 2
    assert particles.shape[1] == 3

    vv.SetData(ns.numpy_to_vtk(np.ascontiguousarray(particles.astype(np.float64)), deep=1))
    vv.SetDataTypeToDouble()

    out.Allocate(1,1)
    out.SetPoints(vv)

    if aux is not None:
      for n,a in aux:
        a_vtk = ns.numpy_to_vtk(
           np.ascontiguousarray(a.astype(np.float64)
             ), 
           deep=1)
        a_vtk.SetName(n)
        out.GetPointData().AddArray(a_vtk)

    out.InsertNextCell(vtk.VTK_VERTEX, particles.shape[0], range(particles.shape[0]))

  @staticmethod
  def request_information(source):
   pass



class ProgrammableParticleHistoryLoad(object):

  @staticmethod
  def main_script(source, particles, velocities=None, aname="default",addtime=False):
    import vtk
    from vtk.util import numpy_support as ns
  
    out = source.GetOutput()
    vv = vtk.vtkPoints()

    assert len(particles.shape) == 3
    assert particles.shape[2] == 3
    
    if not velocities is None:
      for i,j in zip(velocities.shape,particles.shape):
        assert i==j

    Ntime,Npart,_ = particles.shape

    vv.SetData(ns.numpy_to_vtk(np.ascontiguousarray(particles.reshape((Ntime*Npart,3)).astype(np.float64)), deep=1))
    vv.SetDataTypeToDouble()

    out.Allocate(1,1)
    out.SetPoints(vv)

    if not velocities is None:
      print("Adding velocities")
      vel_vtk = ns.numpy_to_vtk(np.ascontiguousarray(velocities.reshape((Ntime*Npart,3)).astype(np.float64)), deep=1)
      vel_vtk.SetName("velocities")
      out.GetPointData().AddArray(vel_vtk)

    if addtime:
      timearray = np.arange(Ntime)[:,None].repeat(Npart, axis=1).reshape(Ntime*Npart)
      timearray = ns.numpy_to_vtk(np.ascontiguousarray(timearray.astype(np.float64)), deep=1)
      timearray.SetName("timearray")
      out.GetPointData().AddArray(timearray)

    out.InsertNextCell(vtk.VTK_VERTEX, particles.shape[0], range(particles.shape[0]))
    for p in range(Npart):
      out.InsertNextCell(vtk.VTK_LINE, Ntime, range(p, p + Npart*Ntime, Npart) )


  @staticmethod
  def request_information(source):
   pass


class ProgrammableDensityLoad(object):

  @staticmethod
  def main_script(source, density, extents=None, aname="default", extra_arrays={}):
    import vtk
    from vtk.util import numpy_support
  
    if len(density.shape) > 3:
      _, Nx, Ny, Nz = density.shape
    else:
      Nx, Ny, Nz = density.shape

    ido = source.GetOutput()
    ido.SetDimensions(Nx, Ny, Nz)
    if not extents is None:
      origin = extents[:6:2]
      spacing = (extents[1]-extents[0])/Nx, (extents[3]-extents[2])/Ny, (extents[5]-extents[4])/Nz
    else:
      origin = (-1, -1, -1)
      spacing = 2.0 / Nx, 2.0/Ny, 2.0/Nz
      
    ido.SetOrigin(*origin)
    ido.SetSpacing(*spacing)
    ido.SetExtent([0,Nx-1,0,Ny-1,0,Nz-1])
    if len(density.shape) > 3 and density.shape[0] == 3:
       N = Nx*Ny*Nz
       density = density.transpose().astype(np.float64).reshape((N,3))
       arr = numpy_support.numpy_to_vtk(density, deep=1)
    else:
       arr = numpy_support.numpy_to_vtk(density.transpose().astype(np.float64).ravel(), deep=1)
    arr.SetName(aname)
    ido.GetPointData().AddArray(arr)

    for k in extra_arrays.keys():
        arr = numpy_support.numpy_to_vtk(extra_arrays[k].transpose().astype(np.float64).ravel(), deep=1)
        arr.SetName(k)
        ido.GetPointData().AddArray(arr)
    
  @staticmethod
  def request_information(source, density=None, dims=None):
    import vtk
    
    Nx = Ny = Nz = None
    if not density is None:
      Nx, Ny, Nz = density.shape
    elif not dims is None:
      Nx, Ny, Nz = dims
    else:
      raise ValueError("Need at least a density or dims")

    source.GetExecutive().GetOutputInformation(0).Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), 0, Nx-1, 0, Ny-1, 0, Nz-1)

  @staticmethod
  def prepare_timesteps_info(algorithm, timesteps):

    def SetOutputTimesteps(algorithm, timesteps):
      executive = algorithm.GetExecutive()
      outInfo = executive.GetOutputInformation(0)
      outInfo.Remove(executive.TIME_STEPS())
      for timestep in timesteps:
        outInfo.Append(executive.TIME_STEPS(), timestep)
      outInfo.Remove(executive.TIME_RANGE())
      outInfo.Append(executive.TIME_RANGE(), timesteps[0])
      outInfo.Append(executive.TIME_RANGE(), timesteps[-1])

    SetOutputTimesteps(algorithm, timesteps)

  @staticmethod
  @contextmanager
  def get_timestep(algorithm):

    def GetUpdateTimestep(algorithm):
      """Returns the requested time value, or None if not present"""
      executive = algorithm.GetExecutive()
      outInfo = executive.GetOutputInformation(0)
      if not outInfo.Has(executive.UPDATE_TIME_STEP()):
          return None
      return outInfo.Get(executive.UPDATE_TIME_STEP())

    # This is the requested time-step. This may not be exactly equal to the 
    # timesteps published in RequestInformation(). Your code must handle that
    # correctly
    req_time = GetUpdateTimestep(algorithm)

    output = algorithm.GetOutput()

    yield req_time
    
    # Now mark the timestep produced.
    output.GetInformation().Set(output.DATA_TIME_STEP(), req_time)

    
def load_borg(pdo, restart_name, mcmc_name, info=False, aname="BORG"):
  import h5py as h5
  
  with h5.File(restart_name) as f:
    N0 = f["/info/scalars/N0"][:]
    N1 = f["/info/scalars/N1"][:]
    N2 = f["/info/scalars/N2"][:]
    L0 = f["/info/scalars/L0"][:]
    L1 = f["/info/scalars/L1"][:]
    L2 = f["/info/scalars/L2"][:]
    c0 = f["/info/scalars/corner0"][:]
    c1 = f["/info/scalars/corner1"][:]
    c2 = f["/info/scalars/corner2"][:]

  if not info:
    with h5.File(mcmc_name) as f:
      d = f["/scalars/BORG_final_density"][:]+1
                
    ProgrammableDensityLoad.main_script(pdo, d, extents=[c0,c0+L0,c1,c1+L1,c2,c2+L2], aname=aname)
  else:
    ProgrammableDensityLoad.request_information(pdo, dims=[N0,N1,N2])
    
    
def load_borg_galaxies(pdo, restart_name, cid=0, info=False, aname="Galaxies"):
  import h5py as h5
  
  with h5.File(restart_name) as f:
    gals = f['/info/galaxy_catalog_%d/galaxies' % cid]
    ra = gals['phi'][:]
    dec = gals['theta'][:]
    r = gals['r'][:]

  if not info:
    x = r * np.cos(ra)*np.cos(dec)
    y = r * np.sin(ra)*np.cos(dec)
    z = r * np.sin(dec)
    parts = np.array([x,y,z]).transpose()
    ProgrammableParticleLoad.main_script(pdo, parts)
